# =============================================================================
# Copyright (c) 2024 by Laszlo Anka. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# =============================================================================
import inspect
from abc import abstractmethod, ABC
from typing import TYPE_CHECKING, cast, Any, Optional

import yaml

from pypz.core.commons.loggers import ContextLoggerInterface, ContextLogger
from pypz.core.commons.parameters import OptionalParameter
from pypz.core.specs.instance import RegisteredInterface, Instance, InstanceGroup
from pypz.core.specs.dtos import OperatorInstanceDTO, OperatorSpecDTO, OperatorConnection, OperatorConnectionSource
from pypz.core.specs.plugin import Plugin, InputPortPlugin, LoggerPlugin, OutputPortPlugin

if TYPE_CHECKING:
    from pypz.core.specs.pipeline import Pipeline


class Operator(Instance[Plugin], InstanceGroup, RegisteredInterface, ABC):
    """
    This class represents the operator instance specs. This class shall be used
    to integrate your processing logic with *pypz*. An operator spec can contain
    plugins as nested instance.

    :param name: name of the instance, if not provided, it will be attempted to deduce from the variable's name
    """

    # ========================= inner logger class =========================

    class Logger(ContextLoggerInterface):
        """
        This is a wrapper class for the logging functionality. It wraps all the
        implementation of the :class:`LoggerPlugin <pypz.core.specs.plugin.LoggerPlugin>`
        and by invoking either method it will invoke the corresponding method of each
        :class:`LoggerPlugin <pypz.core.specs.plugin.LoggerPlugin>`.

        .. note::
           The logger instance is provided to all plugins, so the plugins can call the
           methods of this logger. However, if a :class:`LoggerPlugin <pypz.core.specs.plugin.LoggerPlugin>`
           would invoke the logger methods, it would cause an infinite recursion. This
           is prevented so that in every logger call, the call trace will be analyzed and
           if any instance of a :class:`LoggerPlugin <pypz.core.specs.plugin.LoggerPlugin>`
           is found, then a ``RecursionError`` will be thrown.

        :param logger_plugins: logger plugin instances collected in the operator instance
        """

        def __init__(self, logger_plugins: set[LoggerPlugin]):
            self.__logger_plugins: set[LoggerPlugin] = logger_plugins

        def _error(self, event: Optional[str] = None, context_stack: list[str] = None, *args: Any, **kw: Any) -> Any:
            # TODO - maybe we should not care for recursion so we can avoid the execution cost?
            frame = inspect.currentframe()
            while (frame := frame.f_back) is not None:
                if ("self" in frame.f_locals) and (isinstance(frame.f_locals["self"], LoggerPlugin)):
                    raise RecursionError(
                        "Attempted to call operator logger from LoggerAddon. This causes infinite recursion.")

            for logger_plugin in self.__logger_plugins:
                logger_plugin._error(event, context_stack, *args, **kw)

        def _warning(self, event: Optional[str] = None, context_stack: list[str] = None, *args: Any, **kw: Any) -> Any:
            # TODO - maybe we should not care for recursion so we can avoid the execution cost?
            frame = inspect.currentframe()
            while (frame := frame.f_back) is not None:
                if ("self" in frame.f_locals) and (isinstance(frame.f_locals["self"], LoggerPlugin)):
                    raise RecursionError(
                        "Attempted to call operator logger from LoggerAddon. This causes infinite recursion.")

            for logger_plugin in self.__logger_plugins:
                logger_plugin._warning(event, context_stack, *args, **kw)

        def _info(self, event: Optional[str] = None, context_stack: list[str] = None, *args: Any, **kw: Any) -> Any:
            # TODO - maybe we should not care for recursion so we can avoid the execution cost?
            frame = inspect.currentframe()
            while (frame := frame.f_back) is not None:
                if ("self" in frame.f_locals) and (isinstance(frame.f_locals["self"], LoggerPlugin)):
                    raise RecursionError(
                        "Attempted to call operator logger from LoggerAddon. This causes infinite recursion.")

            for logger_plugin in self.__logger_plugins:
                logger_plugin._info(event, context_stack, *args, **kw)

        def _debug(self, event: Optional[str] = None, context_stack: list[str] = None, *args: Any, **kw: Any) -> Any:
            # TODO - maybe we should not care for recursion so we can avoid the execution cost?
            frame = inspect.currentframe()
            while (frame := frame.f_back) is not None:
                if ("self" in frame.f_locals) and (isinstance(frame.f_locals["self"], LoggerPlugin)):
                    raise RecursionError(
                        "Attempted to call operator logger from LoggerAddon. This causes infinite recursion.")

            for logger_plugin in self.__logger_plugins:
                logger_plugin._debug(event, context_stack, *args, **kw)

    # ========================= parameters =========================

    _operator_image_name = OptionalParameter(str, alt_name="operatorImageName",
                                             description="The image containing the operator's resources. "
                                                         "It will be used mainly by the deployers.")
    _replication_factor = OptionalParameter(int, alt_name="replicationFactor",
                                            description="Determines, how many replicas "
                                                        "shall be created from the original.",
                                            on_update=lambda instance, val: instance.__replicate())

    # ========================= ctor =========================

    def __init__(self,
                 name: str = None,
                 *args, **kwargs):
        super().__init__(name, Plugin, *args, **kwargs)

        self.__replication_origin: Optional[Operator] = self.get_protected().get_reference()
        """
        Reference to the original instance, which was the base for the replication
        """

        self.__replicas: list[Operator] = list()
        """
        List of replica instances
        """

        self._replication_factor: int = 0 if self.get_protected().get_reference() is None else \
            self.get_protected().get_reference().get_replication_factor()
        """
        PARAMETER - The replication factor specifies, how many replicas shall be created along
        the original instance.
        """

        self._operator_image_name: Optional[str] = None
        """
        PARAMETER - Name of the docker image, where the operator can be found. This is optional for the execution,
        but is required at deployment.
        """

        self.__replication_group_index: int = 0 \
            if "replication_group_index" not in kwargs else kwargs["replication_group_index"]
        """
        Specifies the group index, if the operator is replicated.
        """

        self.__logger_plugins: set[LoggerPlugin] = set()
        """
        Collection of logger addons to store for context logging
        """

        self.__logger: ContextLogger = \
            ContextLogger(Operator.Logger(self.__logger_plugins), self.get_full_name())
        """
        Context logger. For more information refer to Operator.Logger class
        """

    # ==================== implementable methods ====================

    @abstractmethod
    def _on_init(self) -> bool:
        """
        This method shall implement the logic to initialize the operation.
        It will be called after services are started and resources are created.

        :return:

        - True, if finished
        - False, if more iteration required (to not block the execution)
        """
        pass

    @abstractmethod
    def _on_running(self) -> Optional[bool]:
        """
        This method shall implement the actual business logic.
        It will be called after the _on_init has successfully finished

        :return:

        - True, if finished
        - False, if more iteration required (to not block the execution)
        - None, if automatically to be determined based on all inputs i.e., if no more input record, then finish
        """
        pass

    @abstractmethod
    def _on_shutdown(self) -> bool:
        """
        This method shall implement the logic to shut down the operation.
        It will be called, after the _on_running has successfully finished.

        :return:

        - True, if finished
        - False, if more iteration required (to not block the execution)
        """
        pass

    # ==================== public methods ====================

    def get_context(self) -> 'Pipeline':
        return cast('Pipeline', super().get_context())

    def get_logger(self) -> ContextLogger:
        return self.__logger

    def get_replication_factor(self):
        return self._replication_factor

    def get_operator_image_name(self):
        return self._operator_image_name

    def get_group_size(self) -> int:
        return self._replication_factor + 1

    def get_group_index(self) -> Optional[int]:
        return self.__replication_group_index

    def get_group_name(self) -> Optional[str]:
        """
        If an Operator instance is replicated, then the replication group name
        is the actual instance name of the original instance.

        :return: the replication group name
        """

        return None if self.__replication_origin is None else self.__replication_origin.get_full_name()

    def get_group_principal(self) -> Optional[Instance]:
        return self.__replication_origin

    def is_principal(self) -> bool:
        return (self.__replication_origin is None) or (self.__replication_origin is self)

    def get_replica(self, replica_id: int) -> 'Operator':
        """
        Returns the replica instance by id. The id is the actual place in the
        replica list, which is ensured during the replica creation.

        :param replica_id: replica id
        :return: replica specified by the id
        """
        return self.__replicas[replica_id]

    def get_replicas(self) -> list['Operator']:
        """
        :return: replica list
        """

        return self.__replicas

    def get_dto(self) -> OperatorInstanceDTO:
        connections = list()
        for instance in self.get_protected().get_nested_instances().values():
            if isinstance(instance, InputPortPlugin):
                for connected_port in instance.get_connected_ports():
                    connections.append(OperatorConnection(
                        inputPortName=instance.get_simple_name(),
                        source=OperatorConnectionSource(
                            instanceName=connected_port.get_context().get_simple_name(),
                            outputPortName=connected_port.get_simple_name())))

        instance_dto = super().get_dto()

        return OperatorInstanceDTO(name=instance_dto.name,
                                   parameters=instance_dto.parameters,
                                   dependsOn=instance_dto.dependsOn,
                                   spec=OperatorSpecDTO(**instance_dto.spec.__dict__),
                                   connections=connections)

    def update(self, source: OperatorInstanceDTO | dict | str) -> None:
        """
        Overridden to allow connection and replica updates.
        """

        if isinstance(source, str):
            instance_dto = OperatorInstanceDTO(**yaml.safe_load(source))
        elif isinstance(source, dict):
            instance_dto = OperatorInstanceDTO(**source)
        elif isinstance(source, OperatorInstanceDTO):
            instance_dto = source
        else:
            raise TypeError(f"Invalid update source type: {type(source)}")

        super().update(instance_dto)

        # Update connections
        # ==================

        # Replicas shall not update connections as those are shared from the original
        if (instance_dto.connections is not None) and (self.get_context() is not None) and self.is_principal():
            for connection in instance_dto.connections:
                if not self.get_protected().has_nested_instance(connection.inputPortName):
                    raise AttributeError(f"[{self.get_full_name()}] Invalid update: InputPort plugin not found "
                                         f"with name '{connection.inputPortName}'")

                if not self.get_context().get_protected().has_nested_instance(connection.source.instanceName):
                    raise AttributeError(f"[{self.get_full_name()}] Invalid update: source instance not found "
                                         f"in pipeline with name '{connection.source.instanceName}'")

                source_instance = self.get_context().get_protected().get_nested_instance(connection.source.instanceName)

                if not source_instance.get_protected().has_nested_instance(connection.source.outputPortName):
                    raise AttributeError(f"[{self.get_full_name()}] Invalid update: OutputPort plugin not found "
                                         f"in source instance '{connection.source.instanceName}' "
                                         f"with name '{connection.source.outputPortName}'")

                input_port_plugin: InputPortPlugin = cast(
                    InputPortPlugin, self.get_protected().get_nested_instance(connection.inputPortName))
                output_port_plugin: OutputPortPlugin = cast(
                    OutputPortPlugin,
                    source_instance.get_protected().get_nested_instance(connection.source.outputPortName))

                input_port_plugin.connect(output_port_plugin)

    def __replicate(self):
        if not self.is_principal():
            return

        if 0 > self._replication_factor:
            raise ValueError(f"Replication factor cannot be negative: {self._replication_factor}")

        difference = self._replication_factor - len(self.__replicas)

        if 0 == difference:
            return

        if 0 < difference:
            if self.__replication_origin is None:
                self.__replication_origin = self
                self.__replication_group_index = 0

            for idx in range(len(self.__replicas), self._replication_factor):
                replica_dto = self.get_dto()
                replica_dto.name = self.get_simple_name() + "_" + str(idx)

                replica = Operator.create_from_dto(replica_dto,
                                                   context=self.get_context(),
                                                   reference=self,
                                                   replication_group_index=idx+1,
                                                   mock_nonexistent=True)

                if self.get_context() is not None:
                    self.get_context().__setattr__(replica.get_simple_name(), replica)

                # We need to store the replicas in the internal replicas list, since
                # that is the only place, where those can be found. Replicas are not
                # direct part of the nested instances
                self.__replicas.append(replica)
        else:
            replicas_to_remove = self.__replicas[difference:]

            for replica in replicas_to_remove:
                self.__replicas.remove(replica)
                if self.get_context() is not None:
                    del self.get_context().get_protected().get_nested_instances()[replica.get_simple_name()]
                    self.get_context().__delattr__(replica.get_simple_name())

            if 0 == len(self.__replicas):
                self.__replication_origin = None
                self.__replication_group_index = 0

    # ==================== protected methods ====================

    def __on_init_finished__(self, *args, **kwargs):
        super().__on_init_finished__(*args, **kwargs)

        # Logger addon handling
        # =====================

        for nested_instance in self.get_protected().get_nested_instances().values():
            if isinstance(nested_instance, LoggerPlugin):
                self.__logger_plugins.add(nested_instance)

    # ========= static methods ==========

    @staticmethod
    def create_from_dto(instance_dto: OperatorInstanceDTO, *args, **kwargs) -> 'Operator':
        return cast(Operator, Instance.create_from_dto(instance_dto, *args, **kwargs))

    @staticmethod
    def create_from_string(source, *args, **kwargs) -> 'Operator':
        return Operator.create_from_dto(OperatorInstanceDTO(**yaml.safe_load(source)), *args, **kwargs)



# =============================================================================
# Copyright (c) 2024 by Laszlo Anka. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# =============================================================================
import time
import traceback
from abc import ABC
from typing import Any, Optional, Type

from pypz.core.channels.io import ChannelReader, ChannelWriter
from pypz.core.channels.status import ChannelStatus
from pypz.core.commons.utils import current_time_millis
from pypz.core.commons.parameters import RequiredParameter, OptionalParameter
from pypz.core.specs.plugin import InputPortPlugin, ResourceHandlerPlugin, OutputPortPlugin, ExtendedPlugin
from pypz.executors.operator.states import StateOperationRunning, StateOperationInit, StateOperationShutdown

ParamKeyChannelLocationConfig = "channelLocation"
ParamKeyChannelConfig = "channelConfig"
ParamKeySequentialModeEnabled = "sequentialModeEnabled"
ParamKeyPortOpenTimeoutMs = "portOpenTimeoutMs"
ParamKeySyncConnectionsOpen = "syncConnectionsOpen"


class ChannelInputPort(InputPortPlugin, ResourceHandlerPlugin, ExtendedPlugin, ABC):
    """
    This class represents an abstract InputPortPlugin, which utilizes the channels
    to realize transfer functionalities. It means that once there is a proper
    channel implementation, a fully functional InputPortPlugin can be made by
    providing that implementation in the abstract new_input_channel() method.

    :param name: name of the instance, if not provided, it will be attempted to deduce from the variable's name
    :param schema: the schema of the port plugin, which will be used to send/retrieve data
    :param group_mode: if set to True, the all the input ports in the group shall receive all messages
    :param channel_reader_type: the type of the channel reader to be used
    """

    # ======================= parameter descriptors =======================

    _channel_location = RequiredParameter(str, alt_name=ParamKeyChannelLocationConfig,
                                          description="Location of the channel resource")
    _channel_config = OptionalParameter(dict, alt_name=ParamKeyChannelConfig,
                                        description="Configuration of the channel as dictionary")
    _sequential_mode_enabled = OptionalParameter(bool, alt_name=ParamKeySequentialModeEnabled,
                                                 description="If set to True, then this port will wait with the "
                                                             "processing start until all the connected output "
                                                             "ports are finished")
    _port_open_timeout_ms = OptionalParameter(int, alt_name=ParamKeyPortOpenTimeoutMs,
                                              description="Specifies, how long the port shall wait for incoming"
                                                          "connections")
    _sync_connections_open = OptionalParameter(bool, alt_name=ParamKeySyncConnectionsOpen,
                                               description="If set to True, the port will wait for every expected"
                                                           "output ports to be connected")

    # ======================= ctor =======================

    def __init__(self,
                 name: str = None,
                 schema: Any = None,
                 group_mode: bool = False,
                 channel_reader_type: Type[ChannelReader] = None,
                 *args, **kwargs):
        super().__init__(name, schema, group_mode, *args, **kwargs)

        self.channel_reader_type: Type[ChannelReader] = channel_reader_type
        """
        Type of the channel reader to be created by this plugin.
        """

        self._channel_reader: Optional[ChannelReader] = None
        """
        The only channel reader maintained by this plugin.
        """

        self._interrupted: bool = False
        """
        Helper flag to register, if the execution has been interrupted.
        """

        self._delete_resources: bool = True
        """
        This flag signalizes, if resource deletion can be performed. It will
        be cleared, if there is an error on OperationRunning state to allow
        process restart.
        """

        self._need_to_check_connections_opened: bool = True
        """
        Helper flag to signalize, if the channel shall be opened. This is necessary
        to maintain state in _on_port_open().
        """

        self._expected_output_count: Optional[int] = None
        """
        During _on_port_open the expected number of connected outputs will be calculated.
        It then will be used in cases, where we need to wait for all outputs.
        """

        # Expected parameters
        # ===================

        self._channel_location = None
        self._channel_config = dict()
        self._sequential_mode_enabled = False
        self._port_open_timeout_ms = 0
        self._port_open_timeout_start = 0
        self._sync_connections_open = False

    # ======================= method implementations =======================

    def _pre_execution(self) -> None:
        if self.channel_reader_type is None:
            raise AttributeError(f"[{self.get_full_name()}] Channel reader type is not specified")

        if self._channel_reader is None:
            channel_name = self.get_full_name() \
                if self.get_group_principal() is None else self.get_group_principal().get_full_name()
            self._channel_reader = self.channel_reader_type(channel_name, self)
            self._channel_reader.set_location(self._channel_location)
            self._channel_reader.invoke_configure_channel(self._channel_config)

    def _post_execution(self) -> None:
        pass

    def _on_resource_creation(self) -> bool:
        if self.is_principal() and (not self._channel_reader.is_resource_created()):
            if self._channel_reader.invoke_resource_creation():
                self._channel_reader.get_logger().debug("Resource created.")
            else:
                self._channel_reader.get_logger().debug("Waiting for resource creation ...")
                return False
        return True

    def _on_resource_deletion(self) -> bool:
        if self.is_principal() and (not self._channel_reader.is_resource_deleted()):
            # Do not delete resources on error
            if not self._delete_resources:
                return True

            if self._channel_reader.invoke_resource_deletion():
                self._channel_reader.get_logger().debug("Resource deleted.")
            else:
                self._channel_reader.get_logger().debug("Waiting for resource deletion ...")
                return False
        return True

    def commit_current_read_offset(self):
        self._channel_reader.invoke_commit_current_read_offset()

    def _on_port_open(self) -> bool:
        if not self._channel_reader.is_channel_open():
            if self._expected_output_count is None:
                self._expected_output_count = sum(output.get_group_size() for output in self.get_connected_ports())

            if self._channel_reader.invoke_open_channel():
                self._channel_reader.get_logger().debug("Channel opened")

                # Resetting, since we reuse this variable at waiting for all inputs to get connected
                self._port_open_timeout_start = 0
            else:
                if 0 < self._port_open_timeout_ms:
                    if 0 == self._port_open_timeout_start:
                        self._port_open_timeout_start = current_time_millis()
                    elif self._port_open_timeout_ms < \
                            (current_time_millis() - self._port_open_timeout_start):
                        raise TimeoutError(f"Timeout exceeded {self._port_open_timeout_ms} [ms]")

                self._channel_reader.get_logger().debug("Waiting for channel to open...")
                return False

        if ChannelReader.NotInitialized == self._channel_reader.get_read_record_offset():
            self._channel_reader.set_initial_record_offset_auto()
            self._channel_reader.get_logger().debug(f"Initialized offset: "
                                                    f"{self._channel_reader.get_read_record_offset()}")

        if not self._channel_reader.is_channel_started():
            self._channel_reader.start_channel()

        self._channel_reader.invoke_sync_status_update()

        # Waiting for outputs to be connected
        # ===================================

        if self._need_to_check_connections_opened:
            if 0 < self._expected_output_count:
                opened_output_count = len(self._channel_reader.retrieve_connected_channel_unique_names(
                    lambda flt: (self.get_group_name() is None) or
                                (flt.get_channel_group_name() != self.get_group_name())
                ))

                if (0 == opened_output_count) or \
                        (self._sync_connections_open and (self._expected_output_count > opened_output_count)):
                    if 0 < self._port_open_timeout_ms:
                        if 0 == self._port_open_timeout_start:
                            self._port_open_timeout_start = current_time_millis()
                        elif self._port_open_timeout_ms < \
                                (current_time_millis() - self._port_open_timeout_start):
                            raise TimeoutError(f"Timeout exceeded {self._port_open_timeout_ms} [ms]")
                    return False

            self._need_to_check_connections_opened = False

        # Waiting for outputs to be finished
        # ==================================
        # Note that normally it does not make sense to slow down the execution by making it sequential,
        # since it is used a couple of times for demo/comparison purposes.
        # Notice that at this point all other relevant checks has been made
        # so there is at least 1 output connected and open

        if self._sequential_mode_enabled:
            if self._channel_reader.is_any_connected_channel_healthy_and_not_stopped_and_not_closed():
                self.get_logger().debug("Waiting for output channel(s) to finish ...")
                return False

        return True

    def retrieve(self) -> Any:
        return self._channel_reader.invoke_read_records()

    def can_retrieve(self) -> bool:
        if not self._channel_reader.is_channel_open():
            raise PermissionError(f"Method allowed only, if channel is open ({self.get_full_name()})")

        finished_output_count = len(self._channel_reader.retrieve_connected_channel_unique_names(
            lambda flt: ((self.get_group_name() is None) or
                         (flt.get_channel_group_name() != self.get_group_name())) and
                        ((not flt.is_channel_healthy()) or flt.is_channel_stopped() or flt.is_channel_closed())
        ))

        return self._channel_reader.has_records() or (self._expected_output_count > finished_output_count)

    def _on_port_close(self) -> bool:
        if self._channel_reader.is_channel_open():
            if self._channel_reader.can_close() and self._channel_reader.invoke_close_channel():
                self._channel_reader.get_logger().debug("Channel closed.")
            else:
                self._channel_reader.get_logger().debug("Waiting for channel to close...")
                return False

        if self.is_principal():
            self.get_logger().debug("Closing port gracefully ...")
            time.sleep(5)
        return True

    def _on_interrupt(self, system_signal: int = None) -> None:
        self.get_logger().warning(f"Interrupted by system signal: {system_signal}")
        self._interrupted = True

    def _on_error(self, source: Any, exception: Exception) -> None:
        if issubclass(source, (StateOperationInit, StateOperationRunning, StateOperationShutdown)):
            self._delete_resources = False

        if self._channel_reader.is_channel_open():
            self._channel_reader.invoke_sync_send_status_message(ChannelStatus.Error)


class ChannelOutputPort(OutputPortPlugin, ResourceHandlerPlugin, ExtendedPlugin, ABC):
    """
    This class represents an abstract OutputPortPlugin, which utilizes the channels
    to realize transfer functionalities. It means that once there is a proper
    channel implementation, a fully functional OutputPortPlugin can be made by
    providing that implementation in the abstract new_output_channel() method.

    :param name: name of the instance, if not provided, it will be attempted to deduce from the variable's name
    :param schema: the schema of the port plugin, which will be used to send/retrieve data
    :param channel_writer_type: the type of the channel writer to be used
    """

    # ======================= parameter descriptors =======================

    _channel_location = RequiredParameter(str, alt_name=ParamKeyChannelLocationConfig,
                                          description="Location of the channel resource")
    _channel_config = OptionalParameter(dict, alt_name=ParamKeyChannelConfig,
                                        description="Configuration of the channel as dictionary")
    _port_open_timeout_ms = OptionalParameter(int, alt_name=ParamKeyPortOpenTimeoutMs,
                                              description="Specifies, how long the port shall wait for incoming"
                                                          "connections")

    # ======================= ctor =======================

    def __init__(self,
                 name: str = None,
                 schema: Any = None,
                 channel_writer_type: Type[ChannelWriter] = None,
                 *args, **kwargs):
        super().__init__(name, schema, *args, **kwargs)

        self.channel_writer_type: Type[ChannelWriter] = channel_writer_type
        """
        Type of the channel writer to be created by this plugin.
        """

        self._channel_writers: Optional[set[ChannelWriter]] = None
        """
        The channel writers to be maintained by this plugin. Note that a channel
        writer will be created for each connection from this output port plugin.
        """

        self._interrupted: bool = False
        """
        Helper flag to register, if the execution has been interrupted.
        """

        self._delete_resources: bool = True
        """
        This flag signalizes, if resource deletion can be performed. It will
        be cleared, if there is an error on OperationRunning state to allow
        process restart.
        """

        self._resource_deletion_errors: set[ChannelWriter] = set()
        """
        Set to store the channels that had errors on resource deletion. It is necessary
        to be able not to reschedule those channels in re-execution of the state.
        """

        self._port_close_errors: set[ChannelWriter] = set()
        """
        Set to store the channels that had errors on port closure. It is necessary
        to be able not to reschedule those channels in re-execution of the state.
        """

        # Expected parameters
        # ===================

        self._channel_location = None
        self._channel_config = dict()
        self._port_open_timeout_ms = 0
        self._port_open_timeout_start = 0

    # ======================= method implementations =======================

    def _pre_execution(self) -> None:
        if self.channel_writer_type is None:
            raise AttributeError(f"[{self.get_full_name()}] Channel writer type is not specified")

        if self._channel_writers is None:
            self._channel_writers = set()
            for connection in self.get_connected_ports():
                new_output_channel = self.channel_writer_type(connection.get_full_name(), self)
                new_output_channel.set_location(self._channel_location)
                new_output_channel.invoke_configure_channel(self._channel_config)
                self._channel_writers.add(new_output_channel)

    def _post_execution(self) -> None:
        pass

    def _on_resource_creation(self) -> bool:
        if not self.is_principal():
            return True

        all_resources_created = True

        for channel in self._channel_writers:
            if not channel.is_resource_created():
                if channel.invoke_resource_creation():
                    channel.get_logger().debug("Resource created.")
                else:
                    channel.get_logger().debug("Waiting for resource creation ...")
                    all_resources_created = False

        if not all_resources_created:
            return False

        if 0 == len(self._channel_writers):
            self.get_logger().warning("No channel reader(s) connected")

        return True

    def _on_resource_deletion(self) -> bool:
        if not self.is_principal():
            return True

        all_resources_deleted = True

        for channel in self._channel_writers:
            if (not channel.is_resource_deleted()) and (channel not in self._resource_deletion_errors):
                try:
                    if self._delete_resources:
                        if channel.invoke_resource_deletion():
                            channel.get_logger().debug("Resource deleted.")
                        else:
                            channel.get_logger().debug("Waiting for resource deletion ...")
                            all_resources_deleted = False
                except:  # noqa: E722
                    # We catch the fact that error occurred, but not interrupting the deletion
                    # process of the other channels
                    self.get_logger().error(traceback.format_exc())
                    self._resource_deletion_errors.add(channel)

        if not all_resources_deleted:
            return False

        if 0 < len(self._resource_deletion_errors):
            raise InterruptedError(f"Error occurred at resource deletion in the following channels: "
                                   f"{[channel.get_channel_name() for channel in self._resource_deletion_errors]}")

        return True

    def _on_port_open(self) -> bool:
        all_channels_opened = True

        for channel in self._channel_writers:
            if not channel.is_channel_open():
                if channel.invoke_open_channel():
                    channel.get_logger().debug("Channel opened.")
                else:
                    channel.get_logger().debug("Waiting for channel open ...")
                    all_channels_opened = False

        if not all_channels_opened:
            if 0 < self._port_open_timeout_ms:
                if 0 == self._port_open_timeout_start:
                    self._port_open_timeout_start = current_time_millis()
                elif self._port_open_timeout_ms < \
                        (current_time_millis() - self._port_open_timeout_start):
                    raise TimeoutError(f"Timeout exceeded {self._port_open_timeout_ms} [ms]")

            return False

        for channel in self._channel_writers:
            if not channel.is_channel_started():
                channel.start_channel()

        return True

    def _on_port_close(self) -> bool:
        all_channels_closed = True

        for channel in self._channel_writers:
            if channel.is_channel_open() and (channel not in self._port_close_errors):
                try:
                    if channel.can_close() and channel.invoke_close_channel():
                        channel.get_logger().debug("Channel closed.")
                    else:
                        all_channels_closed = False
                        channel.get_logger().debug("Waiting for channel to close...")
                except:  # noqa: E722
                    # We catch the fact that error occurred, but not interrupting the shutdown
                    # process of the other channels
                    self.get_logger().error(traceback.format_exc())
                    self._port_close_errors.add(channel)

        if not all_channels_closed:
            return False

        if 0 < len(self._port_close_errors):
            raise InterruptedError(f"Error occurred at port close in the following channels: "
                                   f"{[channel.get_channel_name() for channel in self._port_close_errors]}")

        if self.is_principal():
            self.get_logger().debug("Closing port gracefully ...")
            time.sleep(5)

        return True

    def send(self, data: Any) -> Any:
        for channel in self._channel_writers:
            channel.invoke_write_records(data)

    def _on_interrupt(self, system_signal: int = None) -> None:
        self.get_logger().warning(f"Interrupted by system signal: {system_signal}")
        self._interrupted = True

    def _on_error(self, source: Any, exception: Exception) -> None:
        if issubclass(source, (StateOperationInit, StateOperationRunning, StateOperationShutdown)):
            self._delete_resources = False

        error_channels: set[ChannelWriter] = set()
        for channel in self._channel_writers:
            try:
                if channel.is_channel_open():
                    channel.invoke_sync_send_status_message(ChannelStatus.Error)
            except:  # noqa: E722
                # We catch the fact that error occurred, but not interrupting the error handling
                # process of the other channels
                self.get_logger().error(traceback.format_exc())
                error_channels.add(channel)

        if 0 < len(error_channels):
            raise InterruptedError(f"Error occurred at error handling in the following channels: "
                                   f"{[channel.get_channel_name() for channel in error_channels]}")



# =============================================================================
# Copyright (c) 2024 by Laszlo Anka. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# =============================================================================
from typing import Type, Callable, TypeVar

from pypz.executors.commons import ExecutionMode, ExitCodes
from pypz.core.specs.operator import Operator
from pypz.core.specs.plugin import Plugin
from pypz.core.specs.utils import resolve_dependency_graph

PluginType = TypeVar('PluginType', bound=Plugin)


class ExecutionContext:
    """
    This class is intended to store context information and helper methods for
    the actual execution. The context itself is based on the current Operator.

    :param operator: the actual operator instance to be executed
    :param exec_mode: :class:`ExecutionMode <pypz.executors.commons.ExecutionMode>`
    """

    def __init__(self, operator: Operator, exec_mode: ExecutionMode):
        self.__operator: Operator = operator
        """
        The operator of the context
        """

        self.__exec_mode: ExecutionMode = exec_mode
        """
        The run mode of the execution. For more information refer to the ExecutorRunMode class
        """

        self.__exit_code: ExitCodes = ExitCodes.NoError
        """
        Exit code of the state machine, which can be modified among the states
        """

        self.__plugin_type_registry: dict[Type[Plugin], set[Plugin]] = dict()
        """
        This member holds all the context entities along their implemented interfaces. It allows
        simple type based iteration/execution.
        """

        for nested_instance in self.__operator.get_protected().get_nested_instances().values():
            for spec_class in nested_instance.get_protected().get_spec_classes():
                if spec_class not in self.__plugin_type_registry:
                    self.__plugin_type_registry[spec_class] = set()
                self.__plugin_type_registry[spec_class].add(nested_instance)

        self.__typed_dependency_graphs: dict[Type[Plugin], list[set[Plugin]]] = dict()
        """
        This member holds the nested instances ordered by their resolved dependency list along
        their types. The key holds the instance type, the value is a list of set, where each list
        element represents a dependency level i.e., the 0th element holds the instances w/o
        dependencies, the 1st element holds the instances that are dependent on instances on
        the 0th level and so on.
        """

        for instance_type, instances in self.__plugin_type_registry.items():
            self.__typed_dependency_graphs[instance_type] = resolve_dependency_graph(instances)

    def get_operator(self) -> Operator:
        return self.__operator

    def get_execution_mode(self) -> ExecutionMode:
        return self.__exec_mode

    def get_exit_code(self):
        return self.__exit_code

    def set_exit_code(self, exit_code: ExitCodes):
        self.__exit_code = exit_code

    def get_plugin_instances_by_type(self, plugin_type: Type[PluginType]) -> set[PluginType]:
        return self.__plugin_type_registry[plugin_type] if plugin_type in self.__plugin_type_registry else set()

    def get_dependency_graph_by_type(self, plugin_type: Type[PluginType]) -> list[set[PluginType]]:
        return self.__typed_dependency_graphs[plugin_type] if plugin_type in self.__plugin_type_registry else []

    def for_each_plugin_instances(self, consumer: Callable[[PluginType], None]) -> None:
        for plugin_instance in self.__operator.get_protected().get_nested_instances().values():
            consumer(plugin_instance)

    def for_each_plugin_objects_with_type(self, plugin_type: Type[PluginType],
                                          consumer: Callable[[PluginType], None]) -> None:
        if plugin_type in self.__plugin_type_registry:
            for plugin_instance in self.__plugin_type_registry[plugin_type]:
                consumer(plugin_instance)



# =============================================================================
# Copyright (c) 2024 by Laszlo Anka. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# =============================================================================
from typing import Any

from pypz.core.commons.loggers import DefaultContextLogger
from pypz.core.commons.parameters import OptionalParameter
from pypz.core.specs.plugin import LoggerPlugin


class DefaultLoggerPlugin(LoggerPlugin, DefaultContextLogger):
    """
    This is the default implementation of the :class:`LoggerPlugin <pypz.core.specs.plugin.LoggerPlugin>`
    interface. It actually does not implement anything, but uses the
    :class:`DefaultContextLogger <pypz.core.commons.loggers.DefaultContextLogger>`. This is
    possible, since both implements the
    :class:`ContextLoggerInterface <pypz.core.commons.loggers.ContextLoggerInterface>`.

    :param name: name of the instance, if not provided, it will be attempted to deduce from the variable's name
    """

    _log_level = OptionalParameter(str,
                                   alt_name="logLevel",
                                   on_update=lambda instance, val: None
                                   if val is None else instance.set_log_level(val))

    def __init__(self, name: str = None, *args, **kwargs):
        LoggerPlugin.__init__(self, name, *args, **kwargs)
        DefaultContextLogger.__init__(self, self.get_full_name())

        self._log_level = "INFO"

    def _on_interrupt(self, system_signal: int = None) -> None:
        pass

    def _on_error(self, source: Any, exception: Exception) -> None:
        pass



# =============================================================================
# Copyright (c) 2024 by Laszlo Anka. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# =============================================================================
import logging
from abc import ABC, abstractmethod
from typing import Any, TYPE_CHECKING, cast, Optional

import yaml

from pypz.core.commons.loggers import ContextLoggerInterface, DefaultContextLogger, ContextLogger
from pypz.core.specs.dtos import PluginInstanceDTO, PluginSpecDTO
from pypz.core.specs.instance import RegisteredInterface, Instance, InstanceGroup

if TYPE_CHECKING:
    from pypz.core.specs.operator import Operator


class Plugin(Instance[None], InstanceGroup, RegisteredInterface, ABC):
    """
    This interface has the only purpose to separate the plugin interfaces
    from other interfaces like the Operator. It is necessary to avoid the
    case, where an Operator could be nested into other Operators. All
    plugin interfaces shall extend this one.

    :param name: name of the instance, if not provided, it will be attempted to deduce from the variable's name
    """

    def __init__(self, name: str = None, *args, **kwargs):
        super().__init__(name, None, *args, **kwargs)

        self.__logger: Optional[ContextLogger] = \
            ContextLogger(self.get_context().get_logger(), self.get_full_name()) \
            if self.get_context() is not None else \
            ContextLogger(DefaultContextLogger(self.get_full_name()))
        """
        Context logger, which is the Operator's logger if Operator context existing, otherwise
        it defaults back to the DefaultContextLogger. Note that Plugin without Operator
        context makes only sense in test cases, hence the default log level is set to DEBUG.
        """

        self.__logger.set_log_level(logging.DEBUG)

    # ==================== public methods =======================

    def get_context(self) -> 'Operator':
        return cast('Operator', super().get_context())

    def get_logger(self) -> ContextLogger:
        return self.__logger

    def get_group_size(self) -> int:
        return 1 if self.get_context() is None else self.get_context().get_group_size()

    def get_group_index(self) -> int:
        return 0 if self.get_context() is None else self.get_context().get_group_index()

    def get_group_name(self) -> Optional[str]:
        return None if self.get_group_principal() is None else self.get_group_principal().get_full_name()

    def get_group_principal(self) -> Optional[Instance]:
        if (self.get_context() is None) or (self.get_context().get_group_principal() is None):
            return None
        return self.get_context().get_group_principal().get_protected().get_nested_instance(self.get_simple_name())

    def is_principal(self) -> bool:
        return True if self.get_context() is None else self.get_context().is_principal()

    def get_dto(self) -> PluginInstanceDTO:
        instance_dto = super().get_dto()

        return PluginInstanceDTO(name=instance_dto.name,
                                 parameters=instance_dto.parameters,
                                 dependsOn=instance_dto.dependsOn,
                                 spec=PluginSpecDTO(**instance_dto.spec.__dict__))

    @staticmethod
    def create_from_string(source, *args, **kwargs) -> 'Plugin':
        return Plugin.create_from_dto(PluginInstanceDTO(**yaml.safe_load(source)), *args, **kwargs)

    @staticmethod
    def create_from_dto(instance_dto: 'PluginInstanceDTO', *args, **kwargs) -> 'Plugin':
        return cast(Plugin, Instance.create_from_dto(instance_dto, *args, **kwargs))


class ResourceHandlerPlugin(Plugin, RegisteredInterface, ABC):
    """
    This plugin interface allows to implement resource management related
    functionalities. The respective methods will be called at specific
    times during the execution. Check :ref:`executor` for more information.

    :param name: name of the instance, if not provided, it will be attempted to deduce from the variable's name
    """

    def __init__(self, name: str = None, *args, **kwargs):
        super().__init__(name, None, *args, **kwargs)

    @abstractmethod
    def _on_resource_creation(self) -> bool:
        """
        This method shall implement the logic to create an arbitrary resource
        of any type.

        :return: True succeeded, False if more iteration required (to not block the execution)
        """
        pass

    @abstractmethod
    def _on_resource_deletion(self) -> bool:
        """
        This method shall implement the logic to destroy the created resource.

        :return: True succeeded, False if more iteration required (to not block the execution)
        """
        pass


# ====================================== Port Plugins ======================================


class PortPlugin(Plugin, RegisteredInterface, ABC):
    """
    This plugin interface allows to implement common methods for
    port both input and output port plugins.

    :param name: name of the instance, if not provided, it will be attempted to deduce from the variable's name
    :param schema: the schema of the port plugin, which will be used to send/retrieve data
    """

    def __init__(self, name: Optional[str] = None,
                 schema: Any = None,
                 *args, **kwargs):
        super().__init__(name, *args, **kwargs)

        self.__connected_ports: set[PortPlugin] = self.get_protected().get_reference().__connected_ports \
            if self.get_protected().get_reference() is not None else set()
        """
        This member holds the information about the connected ports, where
        the key is the replication group names and the value is a list of
        connected ports.
        """

        self.__schema: Any = schema
        """
        The port's schema, which is used to identify the format of the data
        sent through the port.
        """

    # ==================== public methods ====================

    def set_schema(self, schema: Any) -> None:
        self.__schema = schema

    def get_schema(self) -> Any:
        return self.__schema

    def get_connected_ports(self) -> set['PortPlugin']:
        return self.__connected_ports

    def connect(self, other_port: 'PortPlugin') -> None:
        # Sanity checks
        # =============

        if isinstance(other_port, type(self)):
            raise TypeError("Invalid connection attempt: ports of same type cannot be connected")

        # Plugins shall have an operator context
        if (self.get_context() is None) or (other_port.get_context() is None):
            raise AttributeError("Invalid port connection attempt: no operator context available")

        # Plugins in the same operator shall not be connected
        if self.get_context() is other_port.get_context():
            raise AttributeError("Invalid port connection attempt. Ports shall have different operator context.")

        # Operators shall have a pipeline context
        if (self.get_context().get_context() is None) or (other_port.get_context().get_context() is None):
            raise AttributeError("Invalid port connection attempt. No pipeline context available.")

        # Operators shall be in the same pipeline context
        if self.get_context().get_context() is not other_port.get_context().get_context():
            raise AttributeError("Invalid port connection attempt. Operators shall be in the same pipeline context.")

        if (self.__schema is not None) and (other_port.__schema is not None) and (self.__schema != other_port.__schema):
            self.get_logger().warning(f"Mismatching schemas. Expected: {self.__schema}; "
                                      f"Provided: {other_port.__schema}")

        self.__connected_ports.add(other_port)
        other_port.__connected_ports.add(self)

    # ==================== overridable methods ====================

    @abstractmethod
    def _on_port_open(self) -> bool:
        """
        This method shall implement the logic to initialize the i/o port functionalities.

        :return: True succeeded, False if more iteration required (to not block the execution)
        """
        pass

    @abstractmethod
    def _on_port_close(self) -> bool:
        """
        This method shall implement the logic to shut down the i/o port functionalities.

        :return: True succeeded, False if more iteration required (to not block the execution)
        """
        pass


class OutputPortPlugin(PortPlugin, RegisteredInterface, ABC):
    """
    This plugin interface allows to implement data transfer output port for an operator.
    Operators can communicate via ports. Different technologies can be implemented
    allowing operators to talk through them.

    :param name: name of the instance, if not provided, it will be attempted to deduce from the variable's name
    :param schema: the schema of the port plugin, which will be used to send/retrieve data
    """

    def __init__(self,
                 name: Optional[str] = None,
                 schema: Optional[Any] = None,
                 *args, **kwargs):
        super().__init__(name, schema, *args, **kwargs)

    @abstractmethod
    def send(self, data: Any) -> Any:
        """
        This method shall implement the logic to send data provided as argument.
        The implementation shall specify the type of the data and the return value.

        :param data: data to be sent
        :return: tbd by the implementation
        """
        pass


class InputPortPlugin(PortPlugin, RegisteredInterface, ABC):
    """
    This plugin interface allows to implement data transfer input port for an operator.
    Operators can communicate via ports. Different technologies can be implemented
    allowing operators to talk through them.

    :param name: name of the instance, if not provided, it will be attempted to deduce from the variable's name
    :param schema: the schema of the port plugin, which will be used to send/retrieve data
    :param group_mode: if set to True, the all the input ports in the group shall receive all messages
    """

    def __init__(self,
                 name: Optional[str] = None,
                 schema: Optional[Any] = None,
                 group_mode: bool = False,
                 *args, **kwargs):
        super().__init__(name, schema, *args, **kwargs)

        self._group_mode: bool = group_mode
        """
        If True, the InputPortPlugin shall receive all records sent to the group
        """

    def is_in_group_mode(self):
        return self._group_mode

    @abstractmethod
    def can_retrieve(self) -> bool:
        """
        This method shall implement the logic to signalize, whether the InputPort is
        still able to retrieve. Unable can mean for example that the OutputPort
        finished writing. This can be then used to terminate reading.

        :return: True if port can retrieve, False if not
        """
        pass

    @abstractmethod
    def retrieve(self) -> Any:
        """
        This method shall implement the logic to retrieve data through the port.

        :return: tbd by the implementation
        """
        pass

    @abstractmethod
    def commit_current_read_offset(self) -> None:
        """
        This method shall implement the logic of committing the current
        read offset based on the technology used.
        """
        pass

# ====================================== Service Plugins ======================================


class ServicePlugin(Plugin, RegisteredInterface, ABC):
    """
    This plugin interface allows to implement a service. Services are special entities
    in the execution, since those are decoupled from the execution life-cycle, hence
    can run in the background. Examples:
    - mounting service, which mounts and unmounts a remote location
    - listener service, which starts a background thread to listen for something
    - etc.

    :param name: name of the instance, if not provided, it will be attempted to deduce from the variable's name
    """

    def __init__(self, name: str = None, *args, **kwargs):
        super().__init__(name, None, *args, **kwargs)

    @abstractmethod
    def _on_service_start(self) -> bool:
        """
        This method shall implement the starting logic of the service. You can consider
        services as example like mounting service, which only mounts a folder to the
        system or a background service, which starts a background thread.

        :return: True if logic finished, False if it needs more iteration
        """
        pass

    @abstractmethod
    def _on_service_shutdown(self) -> bool:
        """
        This method shall implement the logic of stopping the service and clean up the
        residuals. E.g., a mounting service could unmount or a background thread could
        be stopped.
        VERY IMPORTANT NOTE - you must always check, if your service start method has
        been called, because it can be that it is never called, if there was an exception
        raised from other entity. However, the shutdown will be called anyway.

        :return: True if logic finished, False if it needs more iteration
        """
        pass


# ====================================== Addons ======================================


class ExtendedPlugin(Plugin, RegisteredInterface, ABC):
    """
    This interface extends the normal plugin's lifecycle. The methods
    defined span outside the execution context, hence it can be used,
    if you need to perform some action before and after execution.

    :param name: name of the instance, if not provided, it will be attempted to deduce from the variable's name
    """

    def __init__(self, name: str = None, *args, **kwargs):
        super().__init__(name, None, *args, **kwargs)

    @abstractmethod
    def _pre_execution(self) -> None:
        """
        This method will be called before the executor state machine starts. It
        can be used to perform initialization that is required before the execution.
        """
        pass

    @abstractmethod
    def _post_execution(self) -> None:
        """
        This method is called after the executor and its state machine exited, but
        before the program exits. It can be used to perform finalization/shutdown
        logic that is outside the execution context.
        """
        pass


class LoggerPlugin(Plugin, ContextLoggerInterface, RegisteredInterface, ABC):
    """
    This addon interface allows to implement different logging technologies
    to be used during the execution. Notice that the logger methods are
    coming from the ContextLogger class.

    :param name: name of the instance, if not provided, it will be attempted to deduce from the variable's name
    """

    def __init__(self, name: str = None, *args, **kwargs):
        super().__init__(name, None, *args, **kwargs)



# =============================================================================
# Copyright (c) 2024 by Laszlo Anka. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# =============================================================================
from __future__ import annotations
from types import GenericAlias
from typing import Generic, TypeVar, Type, Any, Callable, Optional

from pypz.core.commons.utils import is_type_allowed

allowed_param_types = (str, int, float, set, list, dict, type(None))


ParameterType = TypeVar("ParameterType")


class ExpectedParameter(Generic[ParameterType]):
    """
    This is a descriptor class, which describes an instance parameter. To do that
    you need to describe the parameter on class level (which shall be one of the
    RegisteredInterface) then you need to refer with the same name as instance variable.
    Usage:

    .. code-block:: python

       class TestImpl(Instance):
           required_param = ExpectedParameter(required=True, parameter_type=str)
           optional_param = ExpectedParameter(required=False, parameter_type=str)

           def __init__(self):
               self.required_param = None
               self.optional_param = "defaultValue"

    This is equivalent to:

    .. code-block:: python

       class TestImpl(Instance):
           required_param = RequiredParameter(str)
           optional_param = OptionalParameter(str)

           def __init__(self):
               self.required_param = None
               self.optional_param = "defaultValue"

    :param required: true, if parameter required, false if not
    :param parameter_type: (str, int, float, set, list, dict, type(None))
    :param alt_name: alternative name for the parameter, if specified it acts as reference to the parameter
    :param on_update: callback to react on value update
    """

    NamePrefix = "__private_instance_parameter__"
    """
    This prefix is used to prefix the actual variables created by this descriptor
    """

    def __init__(self,
                 required: bool,
                 parameter_type: Type[ParameterType],
                 alt_name: Optional[str] = None,
                 description: Optional[str] = None,
                 on_update: Callable[[Any, Any], None] = None):

        if parameter_type is None:
            raise TypeError("Parameter type must be specified")

        if "typing" == parameter_type.__module__:
            raise TypeError(f"Type description not allowed: {parameter_type}")

        self.parameter_type: Type[ParameterType] = parameter_type \
            if not isinstance(parameter_type, GenericAlias) else parameter_type.__origin__  # type: ignore

        self.generic_types: list[type] = None \
            if not isinstance(parameter_type, GenericAlias) else parameter_type.__args__  # type: ignore

        if not issubclass(self.parameter_type, allowed_param_types):
            raise TypeError(f"Invalid parameter type: {self.parameter_type}. Allowed types are: {allowed_param_types}")

        self.required: bool = required

        self.name: Optional[str] = alt_name
        """
        Alternative name for the parameter. It can be used to define the actual
        parameter name with having a different variable name.
        """

        self.description: Optional[str] = description
        """
        Short description of the parameter and its intended usage
        """

        self.on_update: Callable[[Any, Any], None] = on_update
        """
        Callback to execute a specific logic, if the value is changed
        """

    def __set_name__(self, owner, name):
        if name.startswith(f"_{owner.__name__}__"):
            raise TypeError("Parameters cannot be defined for private scope")

        if self.name is None:
            self.name = name

        self.internal_name = ExpectedParameter.NamePrefix + name

    def __get__(self, instance, owner) -> ParameterType:
        if not hasattr(instance, self.internal_name):
            raise AttributeError(f"'{owner.__name__}' has no attribute '{self.name}'")

        return getattr(instance, self.internal_name)

    def __set__(self, instance, value):
        if not is_type_allowed(value, allowed_param_types):
            raise TypeError(f"Invalid parameter value type for '{self.name}': {type(value)}. "
                            f"Allowed types: {allowed_param_types}")

        if (value is not None) and \
                (self.parameter_type is not None) and \
                (not isinstance(value, self.parameter_type)):
            raise TypeError(f"Invalid parameter value type for '{self.name}': {type(value)}. "
                            f"Expected: {self.parameter_type}")

        # TODO - evtl. checking against generic parameters

        if (not hasattr(instance, self.internal_name)) and (self.name not in instance.get_protected().get_parameters()):
            """ This is the normal init path, where an object has just been constructed and
                the parameters' initial value is set """

            setattr(instance, self.internal_name, value)
            instance.get_protected().get_parameters()[self.name] = value
            if self.on_update is not None:
                self.on_update(instance, value)
        elif (not hasattr(instance, self.internal_name)) and (self.name in instance.get_protected().get_parameters()):
            """ This path usually represents the case, where a replica is created. The parameter in the dict is
                already set through the original, but the actual parameter attribute is not yet set in the replica.
                In this case it does not matter, what the value is, since the one in the param dict has precedence. """

            setattr(instance, self.internal_name, instance.get_protected().get_parameters()[self.name])
            if self.on_update is not None:
                self.on_update(instance, value)
        elif (hasattr(instance, self.internal_name)) and (self.name in instance.get_protected().get_parameters()):
            """ This is the normal update path for both the normal and the replicated case. Logic shall be
                executed only, if there is value difference. """

            if getattr(instance, self.internal_name) != value:
                setattr(instance, self.internal_name, value)
                if self.on_update is not None:
                    self.on_update(instance, value)
            if instance.get_protected().get_parameters()[self.name] != value:
                """ Note that if parameter has been set directly on the attribute, then
                    we need to update it in the parameters as well. Notice that
                    we cause recurrence, since this __set__ will be called again. This
                    will not cause any trouble however, since this path is not executed
                    if the new value equals old value, which it does, since we set the
                    value in the just before. Recursion is not nice, but still it is
                    the least complex solution to ensure consistency between parameters
                    and ExpectedParameter attributes. """
                instance.get_protected().get_parameters()[self.name] = value
        else:
            """ This path shall actually never be executed, since there is no case, where the value is set, but
                the parameter dict is not updated. This is ensured by the first path. """
            raise AttributeError("Parameter value is missing from the parameter dict. This should not happen."
                                 "Please contact the devs.")

    def __eq__(self, other: object) -> bool:
        if self is other:
            return True

        return isinstance(other, type(self)) and \
            (self.required == other.required) and \
            (self.parameter_type == other.parameter_type) and \
            (self.name == other.name) and \
            (self.description == other.description) and \
            (self.generic_types == other.generic_types)

    def to_dict(self, instance=None) -> dict:
        return {
            self.name: {
                'type': None if self.parameter_type is None else
                self.parameter_type.__origin__ if isinstance(self.parameter_type, GenericAlias)  # type: ignore
                else self.parameter_type.__name__,
                'required': self.required,
                'description': self.description,
                'currentValue': None if instance is None else self.__get__(instance, None)
            }
        }


class RequiredParameter(ExpectedParameter[ParameterType]):
    """
    Convenience class to represent a required parameter
    Usage:

    .. code-block::

       class TestImpl(Instance):
           required_param = RequiredParameter(str)
           optional_param = OptionalParameter(str)
           def __init__(self):
               self.required_param = None
               self.optional_param = "defaultValue"

    :param parameter_type: (str, int, float, set, list, dict, type(None))
    :param alt_name: alternative name for the parameter, if specified it acts as reference to the parameter
    :param on_update: callback to react on value update
    """

    def __init__(self,
                 parameter_type: Type[ParameterType] = None,
                 alt_name: str = None,
                 description: Optional[str] = None,
                 on_update: Callable[[Any, Any], None] = None):
        super().__init__(True, parameter_type, alt_name, description, on_update)


class OptionalParameter(ExpectedParameter[ParameterType]):
    """
    Convenience class to represent an optional parameter
    Usage:

    .. code-block::

       class TestImpl(Instance):
           required_param = RequiredParameter(str)
           optional_param = OptionalParameter(str)
           def __init__(self):
               self.required_param = None
               self.optional_param = "defaultValue"

    :param parameter_type: (str, int, float, set, list, dict, type(None))
    :param alt_name: alternative name for the parameter, if specified it acts as reference to the parameter
    :param on_update: callback to react on value update
    """

    def __init__(self,
                 parameter_type: Type[ParameterType] = None,
                 alt_name: str = None,
                 description: Optional[str] = None,
                 on_update: Callable[[Any, Any], None] = None):
        super().__init__(False, parameter_type, alt_name, description, on_update)


def retrieve_parameters(input_val, parameter_type: Type[ExpectedParameter]) -> dict[str, ExpectedParameter]:
    """
    This method attempts to retrieve all described parameters with
    either public or protected scope. Private scoped parameters are
    ignored.

    :param input_val: object or class to be mapped
    :param parameter_type: parameter type to look for
    :return: dict of parameter names to parameter objects
    """

    object_class = input_val.__class__ if not isinstance(input_val, type) else input_val

    result: dict[str, ExpectedParameter] = dict()

    for implemented_class in object_class.__mro__:
        for param_name, param in implemented_class.__dict__.items():
            if isinstance(param, parameter_type) and \
                    (not param_name.startswith(f"_{implemented_class.__name__}__")):  # privates ignored
                if param.name in result:
                    raise AttributeError(f"Parameter '{param.name}' is already "
                                         f"declared in '{implemented_class}'")

                result[param.name] = param

    return result



# =============================================================================
# Copyright (c) 2024 by Laszlo Anka. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# =============================================================================
import sys

import yaml

from pypz.core.specs.dtos import PipelineInstanceDTO
from pypz.core.specs.operator import Operator
from pypz.core.specs.pipeline import Pipeline
from pypz.executors.commons import ExecutionMode
from pypz.executors.operator.executor import OperatorExecutor

""" This script is used to invoke operator execution """
if __name__ == "__main__":
    print(sys.argv)
    if 3 > len(sys.argv):
        print("Missing arguments:\n"
              "$1 - path to config file\n"
              "$2 - operator simple name\n"
              f"$3 - [Optional] ExecutionMode (default: {ExecutionMode.Standard.name}). "
              f"Possible values: {[elem.name for elem in ExecutionMode]}",
              file=sys.stderr)
        sys.exit(1)

    with open(sys.argv[1]) as json_file:
        pipeline_dto: PipelineInstanceDTO = PipelineInstanceDTO(**yaml.safe_load(json_file))

        pipeline: Pipeline = Pipeline.create_from_dto(pipeline_dto, mock_nonexistent=True)
        operator: Operator = pipeline.get_protected().get_nested_instance(sys.argv[2])
        exec_mode: ExecutionMode = ExecutionMode.Standard if 4 > len(sys.argv) else ExecutionMode(sys.argv[3])

        print(f"Operator to execute: {operator.get_full_name()}; Execution mode: {exec_mode.name}")

        executor: OperatorExecutor = OperatorExecutor(operator)

        exit_code = executor.execute(exec_mode)

        sys.exit(exit_code)



# =============================================================================
# Copyright (c) 2024 by Laszlo Anka. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# =============================================================================
import types
import re
import typing
from importlib import import_module
from typing import Callable, Any, Iterable

IncludedCascadingParameterPrefix = "#"
"""
This strings denotes a cascading parameter. Included cascading parameter means
that the model defines this parameter will get it provided as well the sub instances.
"""

ExcludedCascadingParameterPrefix = ">"
"""
This strings denotes a cascading parameter. Excluded cascading parameter means
that the model defines this parameter will NOT get it provided ONLY the sub instances.
"""


class InstanceParameters(dict):
    """
    This class represents the parameters of the instance organized into
    a dictionary. The necessity of extending the builtin dict is that
    the parameter changes shall be watched because the expected parameters
    that are represented by instance attributes shall be updated as well,
    so we need to be able to define callbacks on parameter set.
    The mentioned default callbacks are defined in the Instance class.
    """

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.__update_callbacks: dict[str, list[Callable[[Any], None]]] = dict()

    def __setitem__(self, name, value):
        if (name not in self) or (value != self[name]):
            super().__setitem__(name, value)
            if name in self.__update_callbacks:
                for callback in self.__update_callbacks[name]:
                    callback(value)

    @typing.no_type_check
    def update(self, __m, **kwargs) -> None:
        super().update(__m, **kwargs)

        for name, value in __m.items():
            if ((name not in self) or (value != self[name])) and (name in self.__update_callbacks):
                for callback in self.__update_callbacks[name]:
                    callback(value)

    def on_parameter_update(self, name, callback: Callable[[Any], None]):
        if name not in self.__update_callbacks:
            self.__update_callbacks[name] = list()
        self.__update_callbacks[name].append(callback)


def resolve_dependency_graph(instances: set | Iterable) -> list[set]:
    """
    This method can resolve the dependencies along instances provided in the instance set.
    Dependency can be formulated between instances via depends_on(). Resolution means that
    the instances will be ordered into a list of sets, where each outer list represents
    a dependency level and each inner sets hold the instances on that level. For example,
    if there are instances w/o any dependencies, then those will be placed to level 0, which
    is the list[0]. On level 1 all the instances are placed that has dependencies to level 0
    instances and so on.

    :param instances: set of instances to resolve the dependencies across
    :return: list of sets, where the list represents dependency levels and set the instances on it
    """
    instance_set = instances if isinstance(instances, set) else set(instances)

    if 0 < len(instance_set):
        resolved_instances: set = set()
        dependency_level_list: list[set] = list()

        for level in range(len(instance_set)):
            # We cannot have more dependency levels than instances we have, hence
            # the outer loop can always create a dependency level, where the instances
            # on the actual level will be stored
            dependency_level_list.append(set())

            for instance in instance_set:
                # It is possible that instances have dependencies that are not present
                # in the list provided as argument, hence we need to create an intersection,
                # so we attempt to resolve dependencies only present in the provided list.
                available_dependencies = instance.get_protected().get_depends_on().intersection(instance_set)

                if (instance not in resolved_instances) and \
                        ((0 == len(available_dependencies)) or
                         available_dependencies.issubset(resolved_instances)):
                    dependency_level_list[level].add(instance)

            resolved_instances.update(dependency_level_list[level])
            if len(resolved_instances) == len(instance_set):
                # Early termination, since all instance dependencies have been resolved
                return dependency_level_list

        # If resolution could not be concluded in N*N steps, then it is very possible
        # that there is a circular dependency. Notice that the method depends_on()
        # checks for circular dependencies, still there is an edge case, where that
        # list is altered outside of that method.
        raise RecursionError("Circular dependency detected in instance dependencies")

    return []


class SingletonAccessWrapper(type):
    """
    This metaclass extends the AccessWrapper class with the functionality
    of caching AccessWrapper objects based on the Instance object. This
    allows to reuse the AccessWrapper object instead of creating a new one.
    """

    __singletons: dict[int, 'AccessWrapper'] = dict()

    def __call__(cls, instance, *args, **kwargs):
        if id(instance) in SingletonAccessWrapper.__singletons:
            return SingletonAccessWrapper.__singletons[id(instance)]
        instance_access = type.__call__(cls, instance, *args, **kwargs)
        SingletonAccessWrapper.__singletons[id(instance)] = instance_access
        return instance_access


class AccessWrapper(metaclass=SingletonAccessWrapper):
    """
    This class wraps the access of the provided object's protected and private
    attributes into publicly accessible methods.
    """

    PATTERN = r"^(_[a-zA-Z0-9]+__|_(?!_))"

    def __init__(self, instance: object):
        # Wrapping methods
        # ================

        for name in dir(instance):
            if hasattr(instance, name):
                attr = getattr(instance, name)
                if isinstance(attr, types.MethodType) and (not re.match(r"^__\w+__$", name)):
                    public_name = re.sub(AccessWrapper.PATTERN, "", name)
                    if (not hasattr(instance, public_name)) or \
                            (not isinstance(getattr(instance, public_name), types.MethodType)):
                        self.__dict__[public_name] = attr

        # Wrapping fields
        # ===============

        for name, value in instance.__dict__.items():
            object.__setattr__(self, name, value)

            if re.match(AccessWrapper.PATTERN, name):
                getter_name = re.sub(AccessWrapper.PATTERN, "get_", name)
                if (not hasattr(instance, getter_name)) or \
                        (not isinstance(getattr(instance, getter_name), types.MethodType)):
                    self.__dict__[getter_name] = types.MethodType(lambda this, n=name: this.__dict__[n], self)

    def get_nested_instance(self, name):
        return self.get_nested_instances()[name]

    def has_nested_instance(self, name):
        return name in self.get_nested_instances()

    def __getattr__(self, name):
        return self.__getattribute__(name)


def load_class_by_name(class_name: str) -> type:
    """
    This method loads a class given its name by traversing its module path up
    to the class itself.
    """

    if 0 > class_name.find("."):
        raise ValueError(f"Class loading error. Invalid class specifier: {class_name}; "
                         f"Class specifier must be composed of [PACKAGE_NAME.]MODULE_NAME:CLASS_NAME[.CLASS_NAME+]")
    trimmed_name = class_name.replace("<class '", "").replace("'>", "")

    loaded = None
    for segment in trimmed_name.split("."):
        if loaded is None:
            loaded = import_module(segment)
        else:
            if hasattr(loaded, segment):
                loaded = getattr(loaded, segment)
            else:
                try:
                    loaded = import_module(loaded.__name__ + "." + segment)
                except ModuleNotFoundError:
                    raise AttributeError(f"Class not found: {loaded.__name__}.{segment}")

    if not isinstance(loaded, type):
        raise TypeError(f"Specified class name results in a non-type: {loaded}")

    return loaded


def remove_super_classes(classes: set[type]) -> set[type]:
    """
    This function will remove the classes from the set, which are super class for any
    of the others, hence redundant. It is used to clean up the Instance specs to avoid
    Deadly Diamond of Death situation at dynamic class creation.
    """

    return {cls for cls in classes if not any(issubclass(pot, cls) and cls is not pot for pot in classes)}



# =============================================================================
# Copyright (c) 2024 by Laszlo Anka. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# =============================================================================
from __future__ import annotations
import concurrent.futures
import time
from concurrent.futures import ThreadPoolExecutor
from abc import ABC, abstractmethod
from typing import Callable, Type, Any, Optional
import types
import traceback
import inspect

from pypz.core.commons.loggers import ContextLogger
from pypz.executors.commons import ExecutionMode
from pypz.executors.commons import ExitCodes
from pypz.executors.operator.context import ExecutionContext
from pypz.core.commons.utils import current_time_millis
from pypz.core.specs.instance import Instance
from pypz.core.specs.plugin import ResourceHandlerPlugin, ServicePlugin, PortPlugin, InputPortPlugin
from pypz.executors.operator.signals import BaseSignal, SignalNoOp, SignalServicesStart, SignalTerminate, \
    SignalOperationStart, SignalError, SignalOperationStop, SignalServicesStop, SignalResourcesDeletion, \
    SignalOperationInit, SignalResourcesCreation, SignalKill


class State(ABC):
    """
    This class is used to collect and abstract the common logic in all derived states.
    All other states must extend on this class.

    :param context: :class:`pypz.executors.operator.context.ExecutionContext`
    """

    # ============ inner classes =============

    class Execution:
        def __init__(self, method: Any, instances: set[Instance], *args, **kwargs):
            self.method = method
            self.instances = instances
            self.args = args
            self.kwargs = kwargs

    class MethodWrapper:
        """
        This class represents the invocation of the plugin objects' method as action. It
        is used by both the parallel and the sequential executor.

        :param state: the state object, where the execution is performed
        :param instance_name: name of the executable instance
        :param callable_method: the method of the instance to be invoked
        """

        def __init__(self, state: State,
                     instance_name: str,
                     callable_method: Callable[..., bool]):
            self.__state = state
            self.__instance_name = instance_name
            self.__callable_method = callable_method

        def __call__(self, *args, **kwargs):

            before_timestamp_ms = current_time_millis()

            try:
                # Invocation of the actual plugin object's method, which was given as
                # parameter to this class
                is_completed = self.__callable_method(*args, **kwargs)

                # If method's return type has been annotated, then we need to check, if the
                # method indeed returns the expected type. Raising error if not.
                method_annotations = inspect.get_annotations(self.__callable_method)

                # Retrieve the return type. Notice that it might be None in which case we need to set
                # the type to NoneType to allow checking isinstance
                return_type = None if "return" not in method_annotations else \
                    types.NoneType if method_annotations.get("return") is None else method_annotations.get("return")

                if ((return_type is not None) and (return_type is not Optional) and
                        (not isinstance(is_completed, return_type))):
                    raise TypeError(f"Invalid return type: {type(is_completed)}, boolean expected.")
            finally:
                after_timestamp_ms = current_time_millis()
                elapsed_time_ms = after_timestamp_ms - before_timestamp_ms

                self.__state._logger.debug("%s | Method: %s; State: %s; Start timestamp: %s; "
                                           "Elapsed time: %s [ms]",
                                           self.__instance_name, self.__callable_method.__name__,
                                           self.__state.__class__.__name__, before_timestamp_ms,
                                           elapsed_time_ms)

            return is_completed

    # ============ ctor =============

    def __init__(self, context: ExecutionContext, *args, **kwargs):

        self.__executor: ThreadPoolExecutor = ThreadPoolExecutor(thread_name_prefix=self.__class__.__name__) \
            if "executor_pool_size" not in kwargs else \
            ThreadPoolExecutor(kwargs["executor_pool_size"], thread_name_prefix=self.__class__.__name__)
        """
        Handles the execution of plugins' corresponding methods, which can run parallel
        """

        self._transition_map: dict[Type[BaseSignal], State] = dict()
        """
        Contains the possible transitions from this state
        """

        self._reason: Optional[BaseSignal] = None
        """
        Contains the information about the signal that drove the
        state machine into this state i.e., the reason of this state
        """

        self._prev_state: Optional[State] = None
        """
        Contains the information about the previous state the
        state machine was in
        """

        self._context: ExecutionContext = context
        """
        Holds the reference to the actual operator instance that shall be executed
        """

        self._state_entry_time_ms: int = 0
        """
        Stores the state entry epoch time
        """

        self._response_collector: dict[Callable[[], bool], bool] = dict()
        """
        Stores the responses of the plugin objects for each instance. Used to prevent finished
        instances to be re-scheduled
        """

        self._logger: ContextLogger = ContextLogger(self._context.get_operator().get_logger(),
                                                    self._context.get_operator().get_full_name(),
                                                    self.__class__.__name__)

    # ======== abstract methods =========

    @abstractmethod
    def on_entry(self) -> None:
        pass

    @abstractmethod
    def on_exit(self) -> None:
        pass

    @abstractmethod
    def on_execute(self) -> BaseSignal:
        pass

    # ======== public methods =========

    def get_prev_state(self) -> Optional[State]:
        return self._prev_state

    def get_transitions(self):
        return self._transition_map

    def set_transition(self, signal: Type[BaseSignal], new_state: State) -> None:
        self._logger.debug("Setting up transition: (%s)--[%s]-->(%s)", self.__class__.__name__,
                           signal.__name__, new_state.__class__.__name__)

        if signal in self._transition_map:
            raise AttributeError(f"Signal already registered in transition map: {signal}")

        self._transition_map[signal] = new_state

    def on_signal_handling(self, signal: BaseSignal) -> State:
        if not isinstance(signal, BaseSignal):
            raise AttributeError(f"Invalid signal type: {type(signal)}")

        if signal.__class__ in self._transition_map:
            new_state = self._transition_map[signal.__class__]

            new_state._reason = signal
            new_state._prev_state = self

            self.__exit_state()
            new_state.__enter_state()

            return new_state
        else:
            if not isinstance(signal, SignalNoOp):
                self._logger.warning("Unhandled signal in '%s': %s",
                                     self.__class__.__name__, signal.__class__.__name__)

            return self

    def _schedule(self, *execution_chain: Execution,
                  break_on_exception: bool = False) -> bool:
        """
        Schedules a chain of instances' specified methods for execution via the thread pool
        executor. Each chain element is represented by a list of instances and the method to
        be called. Each list will be executed parallel. The next chain element will be
        executed after all instance methods have been concluded on the current chain element.
        This method blocks until all the scheduled methods are finished.
        Both plugin and operator instance methods can be scheduled.

        :param execution_chain: list of tuples, where [0] is the callable method and [1] is the list of instances
        :param break_on_exception: if True, it will break on the first exception, if False, it will let then chain run
                                   and only raise an exception at the end
        :return: True if all methods are finished (returned True), False otherwise
        """

        all_instances_finished: bool = True
        error_in_instances: list[str] = list()

        for execution in execution_chain:

            method_name = execution.method \
                if isinstance(execution.method, str) else execution.method.__name__  # type: ignore

            futures: dict[concurrent.futures.Future, Callable[[], Any]] = dict()
            method_to_instance: dict[Callable[[], Any], Instance] = dict()

            for instance in execution.instances:
                # Early termination in case an actual instance object does not implement the
                # specified method
                if not hasattr(instance, method_name):
                    continue

                method_reference = getattr(instance, method_name)

                if not isinstance(method_reference, types.MethodType):
                    raise TypeError(f"Invalid callable method type: {type(method_reference)}. "
                                    f"Must be FunctionType of Callable[[], bool].")

                method_to_instance[method_reference] = instance

                if method_reference not in self._response_collector:
                    self._response_collector[method_reference] = False

                if not self._response_collector[method_reference]:
                    futures[
                        self.__executor.submit(
                            State.MethodWrapper(self, instance.get_simple_name(), method_reference),
                            *execution.args, **execution.kwargs
                        )
                    ] = method_reference

            for future, method_reference in futures.items():
                # Early termination, if the future has been prematurely cancelled. This can be
                # the case, if a shutdown hook has been caught and the state executors has been
                # shut down in it
                if future.cancelled():
                    continue

                try:
                    self._response_collector[method_reference] = future.result()

                    # With this logic we make sure that once the flag has been set to False
                    # it remains False i.e., if one instance is not complete the entire
                    # chain is not complete
                    if all_instances_finished and isinstance(self._response_collector[method_reference], bool):
                        all_instances_finished = self._response_collector[method_reference]
                except Exception as e:
                    self._logger.error(f"Exception at {method_reference} of "
                                       f"{method_to_instance[method_reference].get_simple_name()}: {e}")
                    self._logger.error(traceback.format_exc())
                    error_in_instances.append(method_to_instance[method_reference].get_simple_name())
                    if break_on_exception:
                        raise RuntimeError(self.__class__.__name__, error_in_instances)

        if 0 < len(error_in_instances):
            raise RuntimeError(self.__class__.__name__, error_in_instances)

        return all_instances_finished

    def shutdown(self):
        self._logger.debug("Shutting down state: %s", self.__class__.__name__)
        self.__executor.shutdown(wait=True, cancel_futures=True)

    # ======== private methods =========

    def __enter_state(self):
        self._logger.info("(%s)--[%s]-->(%s)", self._prev_state.__class__.__name__,
                          self._reason.__class__.__name__, self.__class__.__name__)

        self._response_collector.clear()

        self.on_entry()

        self._state_entry_time_ms = current_time_millis()

    def __exit_state(self):
        self.on_exit()


class StateEntry(State):
    """
    Serves as entry point for the state machine.

    :param context: :class:`pypz.executors.operator.context.ExecutionContext`
    """

    def __init__(self, context: ExecutionContext, *args, **kwargs):
        super().__init__(context, *args, **kwargs)

    def on_entry(self) -> None:
        pass

    def on_exit(self) -> None:
        pass

    def on_execute(self) -> BaseSignal:
        return SignalServicesStart()


class StateKilled(State):
    """
    Serves as exit point for the state machine.

    :param context: :class:`pypz.executors.operator.context.ExecutionContext`
    """

    def __init__(self, context: ExecutionContext, *args, **kwargs):
        super().__init__(context, *args, **kwargs)

    def on_entry(self) -> None:
        pass

    def on_exit(self) -> None:
        pass

    def on_execute(self) -> BaseSignal:
        return SignalTerminate()


class StateOperationInit(State):
    """
    This state initializes the operator.

    Invoked methods:

    1. :meth:`pypz.core.specs.plugin.PortPlugin._on_port_open`
    2. :meth:`pypz.core.specs.operator.Operator._on_init`

    This order of execution guarantees that the Operator's implementation is already having access
    to the ports in the init phase.


    .. note::
       Note that both the :class:`InputPortPlugin <pypz.core.specs.plugin.InputPortPlugin>` and
       the :class:`OutputPortPlugin <pypz.core.specs.plugin.OutputPortPlugin>` inherits from the
       :class:`PortPlugin <pypz.core.specs.plugin.PortPlugin>`, hence plugins of both types will
       be initialized.

    :param context: :class:`pypz.executors.operator.context.ExecutionContext`
    """

    def __init__(self, context: ExecutionContext, *args, **kwargs):
        super().__init__(context, *args, **kwargs)

    def on_entry(self) -> None:
        pass

    def on_exit(self) -> None:
        pass

    def on_execute(self) -> BaseSignal:
        try:
            if not self._schedule(*[State.Execution("_on_port_open", level)
                                    for level in self._context.get_dependency_graph_by_type(PortPlugin)],
                                  State.Execution("_on_init", {self._context.get_operator()}),
                                  break_on_exception=True):
                time.sleep(1)
                return SignalNoOp()

            return SignalOperationStart()
        except Exception as e:
            self._logger.error("========== Exception at execution ==========")
            self._logger.error(traceback.format_exc())
            self._logger.error("============================================")

            try:
                self._schedule(State.Execution("_on_error",
                                               {self._context.get_operator()},
                                               source=self.__class__, exception=e.__context__),
                               State.Execution("_on_error",
                                               self._context.get_plugin_instances_by_type(PortPlugin),
                                               source=self.__class__, exception=e.__context__))
            except Exception as ex:
                self._logger.error(f"Exception at error handling: {ex}")
                self._logger.error(traceback.format_exc())

            self._context.set_exit_code(ExitCodes.StateOperationInitError)

            return SignalError(e)


class StateOperationRunning(State):
    """
    This state calls the main processing method of the Operator:

    :meth:`pypz.core.specs.operator.Operator._on_running`

    After successful finish of the Operator's method, the offset commit on all the
    :class:`InputPortPlugin <pypz.core.specs.plugin.InputPortPlugin>` will be invoked.

    :meth:`pypz.core.specs.plugin.InputPortPlugin.commit_current_read_offset`

    This ensures that even, if the developer did not commit offsets manually in the
    implementation, offsets will still be committed.

    :param context: :class:`pypz.executors.operator.context.ExecutionContext`
    """

    def __init__(self, context: ExecutionContext, *args, **kwargs):
        super().__init__(context, *args, **kwargs)

    def on_entry(self) -> None:
        pass

    def on_exit(self) -> None:
        pass

    def on_execute(self) -> BaseSignal:
        try:
            is_finished = self._context.get_operator()._on_running()
            # TODO - on_running shall be submitted to the executor as well
            # is_finished = self._schedule(State.Execution("_on_running", {self._context.get_operator()}))
            self._schedule(State.Execution("commit_current_read_offset",
                                           self._context.get_plugin_instances_by_type(InputPortPlugin)))

            # If nothing or None is returned from on_running, then it will be automatically
            # determined, whether to terminate the state or not
            if is_finished is None:
                for input_port_plugin in self._context.get_plugin_instances_by_type(InputPortPlugin):
                    if input_port_plugin.can_retrieve():
                        return SignalNoOp()
            elif not is_finished:
                return SignalNoOp()

            return SignalOperationStop()
        except Exception as e:
            self._logger.error("========== Exception at execution ==========")
            self._logger.error(traceback.format_exc())
            self._logger.error("============================================")

            try:
                self._schedule(State.Execution("_on_error",
                                               {self._context.get_operator()},
                                               source=self.__class__, exception=e.__context__),
                               State.Execution("_on_error",
                                               self._context.get_plugin_instances_by_type(PortPlugin),
                                               source=self.__class__, exception=e.__context__))
            except Exception as ex:
                self._logger.error(f"Exception at error handling: {ex}")
                self._logger.error(traceback.format_exc())

            self._context.set_exit_code(ExitCodes.StateOperationError)

            return SignalError(e)


class StateOperationShutdown(State):
    """
    This state shuts down the operator.

    Invoked methods:

    1. :meth:`pypz.core.specs.operator.Operator._on_shutdown`
    2. :meth:`pypz.core.specs.plugin.PortPlugin._on_port_close`

    This order of execution guarantees that the Operator's implementation is still having access
    to the ports in the shutdown phase.

    .. note::
       Note that both the :class:`InputPortPlugin <pypz.core.specs.plugin.InputPortPlugin>` and
       the :class:`OutputPortPlugin <pypz.core.specs.plugin.OutputPortPlugin>` inherits from the
       :class:`PortPlugin <pypz.core.specs.plugin.PortPlugin>`, hence plugins of both types will
       be shut down.

    :param context: :class:`pypz.executors.operator.context.ExecutionContext`
    """

    def __init__(self, context: ExecutionContext, *args, **kwargs):
        super().__init__(context, *args, **kwargs)

    def on_entry(self) -> None:
        pass

    def on_exit(self) -> None:
        pass

    def on_execute(self) -> BaseSignal:
        try:
            if not self._schedule(State.Execution("_on_shutdown", {self._context.get_operator()}),
                                  *[State.Execution("_on_port_close", level)
                                    for level in reversed(self._context.get_dependency_graph_by_type(PortPlugin))]):
                time.sleep(1)
                return SignalNoOp()

            if ExecutionMode.WithoutResourceDeletion == self._context.get_execution_mode():
                return SignalServicesStop()

            return SignalResourcesDeletion()
        except Exception as e:
            self._logger.error("========== Exception at execution ==========")
            self._logger.error(traceback.format_exc())
            self._logger.error("============================================")

            try:
                self._schedule(State.Execution("_on_error",
                                               {self._context.get_operator()},
                                               source=self.__class__, exception=e.__context__),
                               State.Execution("_on_error",
                                               self._context.get_plugin_instances_by_type(PortPlugin),
                                               source=self.__class__, exception=e.__context__))
            except Exception as ex:
                self._logger.error(f"Exception at error handling: {ex}")
                self._logger.error(traceback.format_exc())

            self._context.set_exit_code(ExitCodes.StateOperationShutdownError)

            return SignalError(e)


class StateResourceCreation(State):
    """
    This state is responsible to invoke the resource creation related methods.

    Invoked methods:

    :meth:`pypz.core.specs.plugin.ResourceHandlerPlugin._on_resource_creation`

    :param context: :class:`pypz.executors.operator.context.ExecutionContext`
    """

    def __init__(self, context: ExecutionContext, *args, **kwargs):
        super().__init__(context, *args, **kwargs)

    def on_entry(self) -> None:
        pass

    def on_exit(self) -> None:
        pass

    def on_execute(self) -> BaseSignal:
        try:
            if not self._schedule(*[State.Execution("_on_resource_creation", level)
                                    for level in self._context.get_dependency_graph_by_type(ResourceHandlerPlugin)],
                                  break_on_exception=True):
                time.sleep(1)
                return SignalNoOp()

            if ExecutionMode.ResourceCreationOnly == self._context.get_execution_mode():
                return SignalServicesStop()

            return SignalOperationInit()
        except Exception as e:
            self._logger.error("========== Exception at execution ==========")
            self._logger.error(traceback.format_exc())
            self._logger.error("============================================")

            try:
                self._schedule(State.Execution("_on_error",
                                               self._context.get_plugin_instances_by_type(ResourceHandlerPlugin),
                                               source=self.__class__, exception=e.__context__))
            except Exception as ex:
                self._logger.error(f"Exception at error handling: {ex}")
                self._logger.error(traceback.format_exc())

            self._context.set_exit_code(ExitCodes.StateResourceCreationError)

            return SignalError(e)


class StateResourceDeletion(State):
    """
    This state is responsible to invoke the resource deletion related methods.

    Invoked methods:

    :meth:`pypz.core.specs.plugin.ResourceHandlerPlugin._on_resource_deletion`

    :param context: :class:`pypz.executors.operator.context.ExecutionContext`
    """

    def __init__(self, context: ExecutionContext, *args, **kwargs):
        super().__init__(context, *args, **kwargs)

    def on_entry(self) -> None:
        pass

    def on_exit(self) -> None:
        pass

    def on_execute(self) -> BaseSignal:
        try:
            if not self._schedule(*[State.Execution("_on_resource_deletion", level) for level in
                                    reversed(self._context.get_dependency_graph_by_type(ResourceHandlerPlugin))]):
                time.sleep(1)
                return SignalNoOp()

            return SignalServicesStop()
        except Exception as e:
            self._logger.error("========== Exception at execution ==========")
            self._logger.error(traceback.format_exc())
            self._logger.error("============================================")

            try:
                self._schedule(State.Execution("_on_error",
                                               self._context.get_plugin_instances_by_type(ResourceHandlerPlugin),
                                               source=self.__class__, exception=e.__context__))
            except Exception as ex:
                self._logger.error(f"Exception at error handling: {ex}")
                self._logger.error(traceback.format_exc())

            self._context.set_exit_code(ExitCodes.StateResourcesDeletionError)

            return SignalError(e)


class StateServiceStart(State):
    """
    This state is responsible to invoke the service start related methods.

    Invoked methods:

    :meth:`pypz.core.specs.plugin.ServicePlugin._on_service_start`

    :param context: :class:`pypz.executors.operator.context.ExecutionContext`
    """

    def __init__(self, context: ExecutionContext, *args, **kwargs):
        super().__init__(context, *args, **kwargs)

    def on_entry(self) -> None:
        pass

    def on_exit(self) -> None:
        pass

    def on_execute(self) -> BaseSignal:
        try:
            if not self._schedule(*[State.Execution("_on_service_start", level)
                                    for level in self._context.get_dependency_graph_by_type(ServicePlugin)],
                                  break_on_exception=True):
                time.sleep(1)
                return SignalNoOp()

            if ExecutionMode.ResourceDeletionOnly == self._context.get_execution_mode():
                return SignalResourcesDeletion()

            return SignalResourcesCreation()
        except Exception as e:
            self._logger.error("========== Exception at execution ==========")
            self._logger.error(traceback.format_exc())
            self._logger.error("============================================")

            try:
                self._schedule(State.Execution("_on_error",
                                               self._context.get_plugin_instances_by_type(ServicePlugin),
                                               source=self.__class__, exception=e.__context__))
            except Exception as ex:
                self._logger.error(f"Exception at error handling: {ex}")
                self._logger.error(traceback.format_exc())

            self._context.set_exit_code(ExitCodes.StateServiceStartError)

            return SignalError(e)


class StateServiceShutdown(State):
    """
    This state is responsible to invoke the service shutdown related methods.

    Invoked methods:

    :meth:`pypz.core.specs.plugin.ServicePlugin._on_service_shutdown`

    :param context: :class:`pypz.executors.operator.context.ExecutionContext`
    """

    def __init__(self, context: ExecutionContext, *args, **kwargs):
        super().__init__(context, *args, **kwargs)

    def on_entry(self) -> None:
        pass

    def on_exit(self) -> None:
        pass

    def on_execute(self) -> BaseSignal:
        try:
            if not self._schedule(*[State.Execution("_on_service_shutdown", level)
                                    for level in reversed(self._context.get_dependency_graph_by_type(ServicePlugin))]):
                time.sleep(1)
                return SignalNoOp()

            return SignalKill()
        except Exception as e:
            self._logger.error("========== Exception at execution ==========")
            self._logger.error(traceback.format_exc())
            self._logger.error("============================================")

            try:
                self._schedule(State.Execution("_on_error",
                                               self._context.get_plugin_instances_by_type(ServicePlugin),
                                               source=self.__class__, exception=e.__context__))
            except Exception as ex:
                self._logger.error(f"Exception at error handling: {ex}")
                self._logger.error(traceback.format_exc())

            self._context.set_exit_code(ExitCodes.StateServiceShutdownError)

            return SignalKill()
